import graphene


class GrapheneMetadataItemDefinition(graphene.ObjectType):
    key = graphene.NonNull(graphene.String)
    value = graphene.NonNull(graphene.String)

    class Meta:
        name = "MetadataItemDefinition"



import sys

from dagster import check
from dagster.cli.workspace.context import WorkspaceRequestContext
from dagster.config.validate import validate_config_from_snap
from dagster.core.host_representation import (
    ExternalExecutionPlan,
    ExternalPipeline,
    PipelineSelector,
    RepositorySelector,
)
from dagster.utils.error import serializable_error_info_from_exc_info
from graphql.execution.base import ResolveInfo

from .utils import UserFacingGraphQLError, capture_error


def get_full_external_pipeline_or_raise(graphene_info, selector):
    from ..schema.errors import GraphenePipelineNotFoundError

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(selector, "selector", PipelineSelector)

    if not graphene_info.context.has_external_pipeline(selector):
        raise UserFacingGraphQLError(GraphenePipelineNotFoundError(selector=selector))

    return graphene_info.context.get_full_external_pipeline(selector)


def get_external_pipeline_or_raise(graphene_info, selector):
    from ..schema.pipelines.pipeline_errors import GrapheneInvalidSubsetError
    from ..schema.pipelines.pipeline import GraphenePipeline

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(selector, "selector", PipelineSelector)

    full_pipeline = get_full_external_pipeline_or_raise(graphene_info, selector)

    if selector.solid_selection is None:
        return full_pipeline

    for solid_name in selector.solid_selection:
        if not full_pipeline.has_solid_invocation(solid_name):
            raise UserFacingGraphQLError(
                GrapheneInvalidSubsetError(
                    message='Solid "{solid_name}" does not exist in "{pipeline_name}"'.format(
                        solid_name=solid_name, pipeline_name=selector.pipeline_name
                    ),
                    pipeline=GraphenePipeline(full_pipeline),
                )
            )

    return get_subset_external_pipeline(graphene_info.context, selector)


def get_subset_external_pipeline(context, selector):
    from ..schema.pipelines.pipeline_errors import GrapheneInvalidSubsetError
    from ..schema.pipelines.pipeline import GraphenePipeline

    check.inst_param(selector, "selector", PipelineSelector)

    repository_location = context.get_repository_location(selector.location_name)
    external_repository = repository_location.get_repository(selector.repository_name)

    try:
        subset_result = repository_location.get_subset_external_pipeline_result(selector)
    except Exception:  # pylint: disable=broad-except
        error_info = serializable_error_info_from_exc_info(sys.exc_info())
        raise UserFacingGraphQLError(
            GrapheneInvalidSubsetError(
                message="{message}{cause_message}".format(
                    message=error_info.message,
                    cause_message="\n{}".format(error_info.cause.message)
                    if error_info.cause
                    else "",
                ),
                pipeline=GraphenePipeline(context.get_full_external_pipeline(selector)),
            )
        )

    return ExternalPipeline(
        subset_result.external_pipeline_data,
        repository_handle=external_repository.handle,
    )


def ensure_valid_config(external_pipeline, mode, run_config):
    from ..schema.pipelines.config import GraphenePipelineConfigValidationInvalid

    check.inst_param(external_pipeline, "external_pipeline", ExternalPipeline)
    check.str_param(mode, "mode")
    # do not type check run_config so that validate_config_from_snap throws

    validated_config = validate_config_from_snap(
        config_schema_snapshot=external_pipeline.config_schema_snapshot,
        config_type_key=external_pipeline.root_config_key_for_mode(mode),
        config_value=run_config,
    )

    if not validated_config.success:

        raise UserFacingGraphQLError(
            GraphenePipelineConfigValidationInvalid.for_validation_errors(
                external_pipeline, validated_config.errors
            )
        )

    return validated_config


def ensure_valid_step_keys(full_external_execution_plan, step_keys):
    from ..schema.errors import GrapheneInvalidStepError

    check.inst_param(
        full_external_execution_plan, "full_external_execution_plan", ExternalExecutionPlan
    )
    check.opt_list_param(step_keys, "step_keys", of_type=str)

    if not step_keys:
        return

    for step_key in step_keys:
        if not full_external_execution_plan.has_step(step_key):
            raise UserFacingGraphQLError(GrapheneInvalidStepError(invalid_step_key=step_key))


def get_external_execution_plan_or_raise(
    graphene_info,
    external_pipeline,
    mode,
    run_config,
    step_keys_to_execute,
    known_state,
):
    full_external_execution_plan = graphene_info.context.get_external_execution_plan(
        external_pipeline=external_pipeline,
        run_config=run_config,
        mode=mode,
        step_keys_to_execute=None,
        known_state=None,
    )

    if not step_keys_to_execute:
        return full_external_execution_plan

    ensure_valid_step_keys(full_external_execution_plan, step_keys_to_execute)

    return graphene_info.context.get_external_execution_plan(
        external_pipeline=external_pipeline,
        run_config=run_config,
        mode=mode,
        step_keys_to_execute=step_keys_to_execute,
        known_state=known_state,
    )


@capture_error
def fetch_repositories(graphene_info):
    from ..schema.external import GrapheneRepository, GrapheneRepositoryConnection

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    return GrapheneRepositoryConnection(
        nodes=[
            GrapheneRepository(repository=repository, repository_location=location)
            for location in graphene_info.context.repository_locations
            for repository in location.get_repositories().values()
        ]
    )


@capture_error
def fetch_repository(graphene_info, repository_selector):
    from ..schema.errors import GrapheneRepositoryNotFoundError
    from ..schema.external import GrapheneRepository

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(repository_selector, "repository_selector", RepositorySelector)

    if graphene_info.context.has_repository_location(repository_selector.location_name):
        repo_loc = graphene_info.context.get_repository_location(repository_selector.location_name)
        if repo_loc.has_repository(repository_selector.repository_name):
            return GrapheneRepository(
                repository=repo_loc.get_repository(repository_selector.repository_name),
                repository_location=repo_loc,
            )

    return GrapheneRepositoryNotFoundError(
        repository_selector.location_name, repository_selector.repository_name
    )


@capture_error
def fetch_workspace(workspace_request_context):
    from ..schema.external import GrapheneWorkspace, GrapheneWorkspaceLocationEntry

    check.inst_param(
        workspace_request_context, "workspace_request_context", WorkspaceRequestContext
    )

    nodes = [
        GrapheneWorkspaceLocationEntry(entry)
        for entry in workspace_request_context.workspace_snapshot.values()
    ]

    return GrapheneWorkspace(locationEntries=nodes)



from dagster import check
from dagster.core.definitions.run_request import JobType
from dagster.core.host_representation import PipelineSelector, RepositorySelector, SensorSelector
from dagster.core.scheduler.job import JobState, JobStatus
from dagster.seven import get_current_datetime_in_utc, get_timestamp_from_utc_datetime
from graphql.execution.base import ResolveInfo

from .utils import UserFacingGraphQLError, capture_error


@capture_error
def get_sensors_or_error(graphene_info, repository_selector):
    from ..schema.sensors import GrapheneSensor, GrapheneSensors

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(repository_selector, "repository_selector", RepositorySelector)

    location = graphene_info.context.get_repository_location(repository_selector.location_name)
    repository = location.get_repository(repository_selector.repository_name)

    return GrapheneSensors(
        results=[
            GrapheneSensor(graphene_info, sensor) for sensor in repository.get_external_sensors()
        ]
    )


@capture_error
def get_sensor_or_error(graphene_info, selector):
    from ..schema.errors import GrapheneSensorNotFoundError
    from ..schema.sensors import GrapheneSensor

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(selector, "selector", SensorSelector)
    location = graphene_info.context.get_repository_location(selector.location_name)
    repository = location.get_repository(selector.repository_name)

    if not repository.has_external_sensor(selector.sensor_name):
        raise UserFacingGraphQLError(GrapheneSensorNotFoundError(selector.sensor_name))
    external_sensor = repository.get_external_sensor(selector.sensor_name)

    return GrapheneSensor(graphene_info, external_sensor)


@capture_error
def start_sensor(graphene_info, sensor_selector):
    from ..schema.errors import GrapheneSensorNotFoundError
    from ..schema.sensors import GrapheneSensor

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(sensor_selector, "sensor_selector", SensorSelector)

    location = graphene_info.context.get_repository_location(sensor_selector.location_name)
    repository = location.get_repository(sensor_selector.repository_name)
    if not repository.has_external_sensor(sensor_selector.sensor_name):
        raise UserFacingGraphQLError(GrapheneSensorNotFoundError(sensor_selector.sensor_name))
    external_sensor = repository.get_external_sensor(sensor_selector.sensor_name)
    graphene_info.context.instance.start_sensor(external_sensor)
    return GrapheneSensor(graphene_info, external_sensor)


@capture_error
def stop_sensor(graphene_info, job_origin_id):
    from ..schema.sensors import GrapheneStopSensorMutationResult

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.str_param(job_origin_id, "job_origin_id")
    instance = graphene_info.context.instance
    job_state = instance.get_job_state(job_origin_id)
    if not job_state:
        return GrapheneStopSensorMutationResult(job_state=None)

    instance.stop_sensor(job_origin_id)
    return GrapheneStopSensorMutationResult(job_state=job_state.with_status(JobStatus.STOPPED))


@capture_error
def get_unloadable_sensor_states_or_error(graphene_info):
    from ..schema.jobs import GrapheneJobState, GrapheneJobStates

    sensor_states = graphene_info.context.instance.all_stored_job_state(job_type=JobType.SENSOR)
    external_sensors = [
        sensor
        for repository_location in graphene_info.context.repository_locations
        for repository in repository_location.get_repositories().values()
        for sensor in repository.get_external_sensors()
    ]

    sensor_origin_ids = {
        external_sensor.get_external_origin_id() for external_sensor in external_sensors
    }

    unloadable_states = [
        sensor_state
        for sensor_state in sensor_states
        if sensor_state.job_origin_id not in sensor_origin_ids
    ]

    return GrapheneJobStates(
        results=[GrapheneJobState(job_state=job_state) for job_state in unloadable_states]
    )


def get_sensors_for_pipeline(graphene_info, pipeline_selector):
    from ..schema.sensors import GrapheneSensor

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(pipeline_selector, "pipeline_selector", PipelineSelector)

    location = graphene_info.context.get_repository_location(pipeline_selector.location_name)
    repository = location.get_repository(pipeline_selector.repository_name)
    external_sensors = repository.get_external_sensors()

    return [
        GrapheneSensor(graphene_info, external_sensor)
        for external_sensor in external_sensors
        if external_sensor.pipeline_name == pipeline_selector.pipeline_name
    ]


def get_sensor_next_tick(graphene_info, sensor_state):
    from ..schema.jobs import GrapheneFutureJobTick

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(sensor_state, "sensor_state", JobState)

    repository_origin = sensor_state.origin.external_repository_origin
    if not graphene_info.context.has_repository_location(
        repository_origin.repository_location_origin.location_name
    ):
        return None

    repository_location = graphene_info.context.get_repository_location(
        repository_origin.repository_location_origin.location_name
    )
    if not repository_location.has_repository(repository_origin.repository_name):
        return None

    repository = repository_location.get_repository(repository_origin.repository_name)
    external_sensor = repository.get_external_sensor(sensor_state.name)

    if sensor_state.status != JobStatus.RUNNING:
        return None

    latest_tick = graphene_info.context.instance.get_latest_job_tick(sensor_state.job_origin_id)
    if not latest_tick:
        return None

    next_timestamp = latest_tick.timestamp + external_sensor.min_interval_seconds
    if next_timestamp < get_timestamp_from_utc_datetime(get_current_datetime_in_utc()):
        return None
    return GrapheneFutureJobTick(sensor_state, next_timestamp)



import sys
import time
from unittest import mock

from dagster import file_relative_path
from dagster.cli.workspace.load import location_origins_from_yaml_paths
from dagster.core.host_representation import ManagedGrpcPythonEnvRepositoryLocationOrigin
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster_graphql.test.utils import execute_dagster_graphql

from .graphql_context_test_suite import GraphQLContextVariant, make_graphql_context_test_suite

WORKSPACE_QUERY = """
query {
   workspaceOrError {
      __typename
      ... on Workspace {
        locationEntries {
          __typename
          id
          name
          locationOrLoadError {
            __typename
            ... on RepositoryLocation {
                id
                name
                repositories {
                    name
                }
                isReloadSupported
            }
            ... on PythonError {
              message
            }
          }
          loadStatus
          displayMetadata {
            key
            value
          }
          updatedTimestamp
        }
      }
      ... on PythonError {
          message
          stack
      }
    }
}
"""


class TestLoadWorkspace(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.non_launchable_in_memory_instance_multi_location()]
    )
):
    def test_load_workspace(self, graphql_context):
        # Add an error origin
        original_origins = location_origins_from_yaml_paths(
            [file_relative_path(__file__, "multi_location.yaml")]
        )
        with mock.patch(
            "dagster.cli.workspace.cli_target.location_origins_from_yaml_paths",
        ) as origins_mock:
            original_origins.append(
                ManagedGrpcPythonEnvRepositoryLocationOrigin(
                    location_name="error_location",
                    loadable_target_origin=LoadableTargetOrigin(
                        python_file="made_up_file.py", executable_path=sys.executable
                    ),
                )
            )

            origins_mock.return_value = original_origins

            reload_time = time.time()

            new_context = graphql_context.reload_workspace()

            result = execute_dagster_graphql(new_context, WORKSPACE_QUERY)

            assert result
            assert result.data
            assert result.data["workspaceOrError"]
            assert result.data["workspaceOrError"]["__typename"] == "Workspace", str(result.data)

            nodes = result.data["workspaceOrError"]["locationEntries"]

            assert len(nodes) == 3

            assert all([node["__typename"] == "WorkspaceLocationEntry" for node in nodes]), str(
                nodes
            )

            success_nodes = [
                node
                for node in nodes
                if node["locationOrLoadError"]["__typename"] == "RepositoryLocation"
            ]
            assert len(success_nodes) == 2

            failures = [
                node for node in nodes if node["locationOrLoadError"]["__typename"] == "PythonError"
            ]
            assert len(failures) == 1
            failure_node = failures[0]

            assert failure_node["name"] == "error_location"
            assert failure_node["loadStatus"] == "LOADED"
            assert "No such file or directory" in failure_node["locationOrLoadError"]["message"]

            for node in nodes:
                assert node["loadStatus"] == "LOADED"
                update_time = node["updatedTimestamp"]
                assert update_time >= reload_time and update_time <= time.time()

                metadatas = node["displayMetadata"]
                metadata_dict = {metadata["key"]: metadata["value"] for metadata in metadatas}

                assert (
                    "python_file" in metadata_dict
                    or "module_name" in metadata_dict
                    or "package_name" in metadata_dict
                )



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestComposites.test_composites[non_launchable_in_memory_instance_managed_grpc_env] 1'] = {
    'pipelineOrError': {
        '__typename': 'Pipeline',
        'name': 'composites_pipeline',
        'solidHandles': [
            {
                'handleID': 'add_four',
                'solid': {
                    'definition': {
                        'inputMappings': [
                            {
                                'definition': {
                                    'name': 'num'
                                },
                                'mappedInput': {
                                    'definition': {
                                        'name': 'num'
                                    },
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            }
                        ],
                        'outputMappings': [
                            {
                                'definition': {
                                    'name': 'result'
                                },
                                'mappedOutput': {
                                    'definition': {
                                        'name': 'result'
                                    },
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            }
                        ],
                        'solids': [
                            {
                                'name': 'adder_1'
                            },
                            {
                                'name': 'adder_2'
                            }
                        ]
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'add_four',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'div_four'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_1',
                'solid': {
                    'definition': {
                        'inputMappings': [
                            {
                                'definition': {
                                    'name': 'num'
                                },
                                'mappedInput': {
                                    'definition': {
                                        'name': 'num'
                                    },
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            }
                        ],
                        'outputMappings': [
                            {
                                'definition': {
                                    'name': 'result'
                                },
                                'mappedOutput': {
                                    'definition': {
                                        'name': 'result'
                                    },
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            }
                        ],
                        'solids': [
                            {
                                'name': 'adder_1'
                            },
                            {
                                'name': 'adder_2'
                            }
                        ]
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'adder_1',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_1.adder_1',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'adder_1',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_1.adder_2',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'adder_2',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_2',
                'solid': {
                    'definition': {
                        'inputMappings': [
                            {
                                'definition': {
                                    'name': 'num'
                                },
                                'mappedInput': {
                                    'definition': {
                                        'name': 'num'
                                    },
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            }
                        ],
                        'outputMappings': [
                            {
                                'definition': {
                                    'name': 'result'
                                },
                                'mappedOutput': {
                                    'definition': {
                                        'name': 'result'
                                    },
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            }
                        ],
                        'solids': [
                            {
                                'name': 'adder_1'
                            },
                            {
                                'name': 'adder_2'
                            }
                        ]
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'adder_2',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_2.adder_1',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'adder_1',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_2.adder_2',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'adder_2',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'div_four',
                'solid': {
                    'definition': {
                        'inputMappings': [
                            {
                                'definition': {
                                    'name': 'num'
                                },
                                'mappedInput': {
                                    'definition': {
                                        'name': 'num'
                                    },
                                    'solid': {
                                        'name': 'div_1'
                                    }
                                }
                            }
                        ],
                        'outputMappings': [
                            {
                                'definition': {
                                    'name': 'result'
                                },
                                'mappedOutput': {
                                    'definition': {
                                        'name': 'result'
                                    },
                                    'solid': {
                                        'name': 'div_2'
                                    }
                                }
                            }
                        ],
                        'solids': [
                            {
                                'name': 'div_1'
                            },
                            {
                                'name': 'div_2'
                            }
                        ]
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'add_four'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'div_four',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'div_four.div_1',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'div_1',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'div_2'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'div_four.div_2',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'div_1'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'div_2',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            }
        ]
    }
}

snapshots['TestComposites.test_composites[non_launchable_sqlite_instance_lazy_repository] 1'] = {
    'pipelineOrError': {
        '__typename': 'Pipeline',
        'name': 'composites_pipeline',
        'solidHandles': [
            {
                'handleID': 'add_four',
                'solid': {
                    'definition': {
                        'inputMappings': [
                            {
                                'definition': {
                                    'name': 'num'
                                },
                                'mappedInput': {
                                    'definition': {
                                        'name': 'num'
                                    },
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            }
                        ],
                        'outputMappings': [
                            {
                                'definition': {
                                    'name': 'result'
                                },
                                'mappedOutput': {
                                    'definition': {
                                        'name': 'result'
                                    },
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            }
                        ],
                        'solids': [
                            {
                                'name': 'adder_1'
                            },
                            {
                                'name': 'adder_2'
                            }
                        ]
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'add_four',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'div_four'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_1',
                'solid': {
                    'definition': {
                        'inputMappings': [
                            {
                                'definition': {
                                    'name': 'num'
                                },
                                'mappedInput': {
                                    'definition': {
                                        'name': 'num'
                                    },
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            }
                        ],
                        'outputMappings': [
                            {
                                'definition': {
                                    'name': 'result'
                                },
                                'mappedOutput': {
                                    'definition': {
                                        'name': 'result'
                                    },
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            }
                        ],
                        'solids': [
                            {
                                'name': 'adder_1'
                            },
                            {
                                'name': 'adder_2'
                            }
                        ]
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'adder_1',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_1.adder_1',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'adder_1',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_1.adder_2',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'adder_2',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_2',
                'solid': {
                    'definition': {
                        'inputMappings': [
                            {
                                'definition': {
                                    'name': 'num'
                                },
                                'mappedInput': {
                                    'definition': {
                                        'name': 'num'
                                    },
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            }
                        ],
                        'outputMappings': [
                            {
                                'definition': {
                                    'name': 'result'
                                },
                                'mappedOutput': {
                                    'definition': {
                                        'name': 'result'
                                    },
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            }
                        ],
                        'solids': [
                            {
                                'name': 'adder_1'
                            },
                            {
                                'name': 'adder_2'
                            }
                        ]
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'adder_2',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_2.adder_1',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'adder_1',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'adder_2'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'add_four.adder_2.adder_2',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'adder_1'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'adder_2',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'div_four',
                'solid': {
                    'definition': {
                        'inputMappings': [
                            {
                                'definition': {
                                    'name': 'num'
                                },
                                'mappedInput': {
                                    'definition': {
                                        'name': 'num'
                                    },
                                    'solid': {
                                        'name': 'div_1'
                                    }
                                }
                            }
                        ],
                        'outputMappings': [
                            {
                                'definition': {
                                    'name': 'result'
                                },
                                'mappedOutput': {
                                    'definition': {
                                        'name': 'result'
                                    },
                                    'solid': {
                                        'name': 'div_2'
                                    }
                                }
                            }
                        ],
                        'solids': [
                            {
                                'name': 'div_1'
                            },
                            {
                                'name': 'div_2'
                            }
                        ]
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'add_four'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'div_four',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'div_four.div_1',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                            ]
                        }
                    ],
                    'name': 'div_1',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                                {
                                    'solid': {
                                        'name': 'div_2'
                                    }
                                }
                            ]
                        }
                    ]
                }
            },
            {
                'handleID': 'div_four.div_2',
                'solid': {
                    'definition': {
                    },
                    'inputs': [
                        {
                            'definition': {
                                'name': 'num'
                            },
                            'dependsOn': [
                                {
                                    'solid': {
                                        'name': 'div_1'
                                    }
                                }
                            ]
                        }
                    ],
                    'name': 'div_2',
                    'outputs': [
                        {
                            'definition': {
                                'name': 'result'
                            },
                            'dependedBy': [
                            ]
                        }
                    ]
                }
            }
        ]
    }
}



import graphene
from dagster import check
from dagster.core.definitions import SolidHandle
from dagster.core.host_representation import RepresentedPipeline
from dagster.core.snap import CompositeSolidDefSnap, DependencyStructureIndex, SolidDefSnap

from .config_types import GrapheneConfigTypeField
from .dagster_types import GrapheneDagsterType, to_dagster_type
from .metadata import GrapheneMetadataItemDefinition
from .util import non_null_list


class _ArgNotPresentSentinel:
    pass


class GrapheneInputDefinition(graphene.ObjectType):
    solid_definition = graphene.NonNull(lambda: GrapheneSolidDefinition)
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    type = graphene.NonNull(GrapheneDagsterType)

    class Meta:
        name = "InputDefinition"

    def __init__(self, represented_pipeline, solid_def_name, input_def_name):
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )
        check.str_param(solid_def_name, "solid_def_name")
        check.str_param(input_def_name, "input_def_name")
        solid_def_snap = self._represented_pipeline.get_solid_def_snap(solid_def_name)
        self._input_def_snap = solid_def_snap.get_input_snap(input_def_name)
        super().__init__(
            name=self._input_def_snap.name,
            description=self._input_def_snap.description,
        )

    def resolve_type(self, _graphene_info):
        return to_dagster_type(
            self._represented_pipeline.pipeline_snapshot, self._input_def_snap.dagster_type_key
        )

    def resolve_solid_definition(self, _graphene_info):
        return build_solid_definition(
            self._represented_pipeline, self._solid_def_snap.name  # pylint: disable=no-member
        )


class GrapheneOutputDefinition(graphene.ObjectType):
    solid_definition = graphene.NonNull(lambda: GrapheneSolidDefinition)
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    is_dynamic = graphene.Boolean()
    type = graphene.NonNull(GrapheneDagsterType)

    class Meta:
        name = "OutputDefinition"

    def __init__(self, represented_pipeline, solid_def_name, output_def_name, is_dynamic):
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )
        check.str_param(solid_def_name, "solid_def_name")
        check.str_param(output_def_name, "output_def_name")

        self._solid_def_snap = represented_pipeline.get_solid_def_snap(solid_def_name)
        self._output_def_snap = self._solid_def_snap.get_output_snap(output_def_name)

        super().__init__(
            name=self._output_def_snap.name,
            description=self._output_def_snap.description,
            is_dynamic=is_dynamic,
        )

    def resolve_type(self, _graphene_info):
        return to_dagster_type(
            self._represented_pipeline.pipeline_snapshot,
            self._output_def_snap.dagster_type_key,
        )

    def resolve_solid_definition(self, _graphene_info):
        return build_solid_definition(self._represented_pipeline, self._solid_def_snap.name)


class GrapheneInput(graphene.ObjectType):
    solid = graphene.NonNull(lambda: GrapheneSolid)
    definition = graphene.NonNull(GrapheneInputDefinition)
    depends_on = non_null_list(lambda: GrapheneOutput)
    is_dynamic_collect = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "Input"

    def __init__(self, represented_pipeline, current_dep_structure, solid_name, input_name):
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )
        self._current_dep_structure = check.inst_param(
            current_dep_structure, "current_dep_structure", DependencyStructureIndex
        )
        self._solid_name = check.str_param(solid_name, "solid_name")
        self._input_name = check.str_param(input_name, "input_name")
        self._solid_invocation_snap = current_dep_structure.get_invocation(solid_name)
        self._solid_def_snap = represented_pipeline.get_solid_def_snap(
            self._solid_invocation_snap.solid_def_name
        )
        self._input_def_snap = self._solid_def_snap.get_input_snap(input_name)

        super().__init__()

    def resolve_solid(self, _graphene_info):
        return GrapheneSolid(
            self._represented_pipeline, self._solid_name, self._current_dep_structure
        )

    def resolve_definition(self, _graphene_info):
        return GrapheneInputDefinition(
            self._represented_pipeline,
            self._solid_def_snap.name,
            self._input_def_snap.name,
        )

    def resolve_depends_on(self, _graphene_info):
        return [
            GrapheneOutput(
                self._represented_pipeline,
                self._current_dep_structure,
                output_handle_snap.solid_name,
                output_handle_snap.output_name,
            )
            for output_handle_snap in self._current_dep_structure.get_upstream_outputs(
                self._solid_name, self._input_name
            )
        ]

    def resolve_is_dynamic_collect(self, _graphene_info):
        return self._solid_invocation_snap.input_dep_snap(self._input_name).is_dynamic_collect


class GrapheneOutput(graphene.ObjectType):
    solid = graphene.NonNull(lambda: GrapheneSolid)
    definition = graphene.NonNull(GrapheneOutputDefinition)
    depended_by = non_null_list(GrapheneInput)

    class Meta:
        name = "Output"

    def __init__(self, represented_pipeline, current_dep_structure, solid_name, output_name):
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )
        self._current_dep_structure = check.inst_param(
            current_dep_structure, "current_dep_structure", DependencyStructureIndex
        )
        self._solid_name = check.str_param(solid_name, "solid_name")
        self._output_name = check.str_param(output_name, "output_name")
        self._solid_invocation_snap = current_dep_structure.get_invocation(solid_name)
        self._solid_def_snap = represented_pipeline.get_solid_def_snap(
            self._solid_invocation_snap.solid_def_name
        )
        self._output_def_snap = self._solid_def_snap.get_output_snap(output_name)
        super().__init__()

    def resolve_solid(self, _):
        return GrapheneSolid(
            self._represented_pipeline, self._solid_name, self._current_dep_structure
        )

    def resolve_definition(self, _graphene_info):
        return GrapheneOutputDefinition(
            self._represented_pipeline,
            self._solid_def_snap.name,
            self._output_name,
            self._output_def_snap.is_dynamic,
        )

    def resolve_depended_by(self, _graphene_info):
        return [
            GrapheneInput(
                self._represented_pipeline,
                self._current_dep_structure,
                input_handle_snap.solid_name,
                input_handle_snap.input_name,
            )
            for input_handle_snap in self._current_dep_structure.get_downstream_inputs(
                self._solid_name, self._output_def_snap.name
            )
        ]


class GrapheneInputMapping(graphene.ObjectType):
    mapped_input = graphene.NonNull(GrapheneInput)
    definition = graphene.NonNull(GrapheneInputDefinition)

    class Meta:
        name = "InputMapping"

    def __init__(
        self,
        represented_pipeline,
        current_dep_index,
        solid_def_name,
        input_name,
    ):
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )
        self._current_dep_index = check.inst_param(
            current_dep_index, "current_dep_index", DependencyStructureIndex
        )
        self._solid_def_snap = represented_pipeline.get_solid_def_snap(solid_def_name)
        self._input_mapping_snap = self._solid_def_snap.get_input_mapping_snap(input_name)
        super().__init__()

    def resolve_mapped_input(self, _graphene_info):
        return GrapheneInput(
            self._represented_pipeline,
            self._current_dep_index,
            self._input_mapping_snap.mapped_solid_name,
            self._input_mapping_snap.mapped_input_name,
        )

    def resolve_definition(self, _graphene_info):
        return GrapheneInputDefinition(
            self._represented_pipeline,
            self._solid_def_snap.name,
            self._input_mapping_snap.external_input_name,
        )


class GrapheneOutputMapping(graphene.ObjectType):
    mapped_output = graphene.NonNull(GrapheneOutput)
    definition = graphene.NonNull(GrapheneOutputDefinition)

    class Meta:
        name = "OutputMapping"

    def __init__(
        self,
        represented_pipeline,
        current_dep_index,
        solid_def_name,
        output_name,
    ):
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )
        self._current_dep_index = check.inst_param(
            current_dep_index, "current_dep_index", DependencyStructureIndex
        )
        self._solid_def_snap = represented_pipeline.get_solid_def_snap(solid_def_name)
        self._output_mapping_snap = self._solid_def_snap.get_output_mapping_snap(output_name)
        self._output_def_snap = self._solid_def_snap.get_output_snap(output_name)

        super().__init__()

    def resolve_mapped_output(self, _graphene_info):
        return GrapheneOutput(
            self._represented_pipeline,
            self._current_dep_index,
            self._output_mapping_snap.mapped_solid_name,
            self._output_mapping_snap.mapped_output_name,
        )

    def resolve_definition(self, _graphene_info):
        return GrapheneOutputDefinition(
            self._represented_pipeline,
            self._solid_def_snap.name,
            self._output_mapping_snap.external_output_name,
            self._output_def_snap.is_dynamic,
        )


class GrapheneResourceRequirement(graphene.ObjectType):
    resource_key = graphene.NonNull(graphene.String)

    class Meta:
        name = "ResourceRequirement"

    def __init__(self, resource_key):
        super().__init__()
        self.resource_key = resource_key


def build_solids(represented_pipeline, current_dep_index):
    check.inst_param(represented_pipeline, "represented_pipeline", RepresentedPipeline)
    return sorted(
        [
            GrapheneSolid(represented_pipeline, solid_name, current_dep_index)
            for solid_name in current_dep_index.solid_invocation_names
        ],
        key=lambda solid: solid.name,
    )


def build_solid_handles(represented_pipeline, current_dep_index, parent=None):
    check.inst_param(represented_pipeline, "represented_pipeline", RepresentedPipeline)
    check.opt_inst_param(parent, "parent", GrapheneSolidHandle)
    all_handle = []
    for solid_invocation in current_dep_index.solid_invocations:
        solid_name, solid_def_name = solid_invocation.solid_name, solid_invocation.solid_def_name
        handle = GrapheneSolidHandle(
            solid=GrapheneSolid(represented_pipeline, solid_name, current_dep_index),
            handle=SolidHandle(solid_name, parent.handleID if parent else None),
            parent=parent if parent else None,
        )
        solid_def_snap = represented_pipeline.get_solid_def_snap(solid_def_name)
        if isinstance(solid_def_snap, CompositeSolidDefSnap):
            all_handle += build_solid_handles(
                represented_pipeline,
                represented_pipeline.get_dep_structure_index(solid_def_name),
                handle,
            )

        all_handle.append(handle)

    return all_handle


class GrapheneISolidDefinition(graphene.Interface):
    name = graphene.NonNull(graphene.String)
    description = graphene.String()
    metadata = non_null_list(GrapheneMetadataItemDefinition)
    input_definitions = non_null_list(GrapheneInputDefinition)
    output_definitions = non_null_list(GrapheneOutputDefinition)

    class Meta:
        name = "ISolidDefinition"


class ISolidDefinitionMixin:
    def __init__(self, represented_pipeline, solid_def_name):
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )
        check.str_param(solid_def_name, "solid_def_name")
        self._solid_def_snap = represented_pipeline.get_solid_def_snap(solid_def_name)

    def resolve_metadata(self, _graphene_info):
        return [
            GrapheneMetadataItemDefinition(key=item[0], value=item[1])
            for item in self._solid_def_snap.tags.items()
        ]

    @property
    def solid_def_name(self):
        return self._solid_def_snap.name

    def resolve_input_definitions(self, _graphene_info):
        return [
            GrapheneInputDefinition(
                self._represented_pipeline, self.solid_def_name, input_def_snap.name
            )
            for input_def_snap in self._solid_def_snap.input_def_snaps
        ]

    def resolve_output_definitions(self, _graphene_info):
        return [
            GrapheneOutputDefinition(
                self._represented_pipeline,
                self.solid_def_name,
                output_def_snap.name,
                output_def_snap.is_dynamic,
            )
            for output_def_snap in self._solid_def_snap.output_def_snaps
        ]


class GrapheneSolidDefinition(graphene.ObjectType, ISolidDefinitionMixin):
    config_field = graphene.Field(GrapheneConfigTypeField)
    required_resources = non_null_list(GrapheneResourceRequirement)

    class Meta:
        interfaces = (GrapheneISolidDefinition,)
        name = "SolidDefinition"

    def __init__(self, represented_pipeline, solid_def_name):
        check.inst_param(represented_pipeline, "represented_pipeline", RepresentedPipeline)
        self._solid_def_snap = represented_pipeline.get_solid_def_snap(solid_def_name)
        super().__init__(name=solid_def_name, description=self._solid_def_snap.description)
        ISolidDefinitionMixin.__init__(self, represented_pipeline, solid_def_name)

    def resolve_config_field(self, _graphene_info):
        return (
            GrapheneConfigTypeField(
                config_schema_snapshot=self._represented_pipeline.config_schema_snapshot,
                field_snap=self._solid_def_snap.config_field_snap,
            )
            if self._solid_def_snap.config_field_snap
            else None
        )

    def resolve_required_resources(self, _graphene_info):
        return [
            GrapheneResourceRequirement(key) for key in self._solid_def_snap.required_resource_keys
        ]


class GrapheneSolid(graphene.ObjectType):
    name = graphene.NonNull(graphene.String)
    definition = graphene.NonNull(lambda: GrapheneISolidDefinition)
    inputs = non_null_list(GrapheneInput)
    outputs = non_null_list(GrapheneOutput)
    is_dynamic_mapped = graphene.NonNull(graphene.Boolean)

    class Meta:
        name = "Solid"

    def __init__(self, represented_pipeline, solid_name, current_dep_structure):
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )

        check.str_param(solid_name, "solid_name")
        self._current_dep_structure = check.inst_param(
            current_dep_structure, "current_dep_structure", DependencyStructureIndex
        )
        self._solid_invocation_snap = current_dep_structure.get_invocation(solid_name)
        self._solid_def_snap = represented_pipeline.get_solid_def_snap(
            self._solid_invocation_snap.solid_def_name
        )
        super().__init__(name=solid_name)

    def get_solid_definition_name(self):
        return self._solid_def_snap.name

    def get_solid_definition(self):
        return build_solid_definition(self._represented_pipeline, self._solid_def_snap.name)

    def resolve_definition(self, _graphene_info):
        return self.get_solid_definition()

    def resolve_inputs(self, _graphene_info):
        return [
            GrapheneInput(
                self._represented_pipeline,
                self._current_dep_structure,
                self._solid_invocation_snap.solid_name,
                input_def_snap.name,
            )
            for input_def_snap in self._solid_def_snap.input_def_snaps
        ]

    def resolve_outputs(self, _graphene_info):
        return [
            GrapheneOutput(
                self._represented_pipeline,
                self._current_dep_structure,
                self._solid_invocation_snap.solid_name,
                output_def_snap.name,
            )
            for output_def_snap in self._solid_def_snap.output_def_snaps
        ]

    def resolve_is_dynamic_mapped(self, _graphene_info):
        return self._solid_invocation_snap.is_dynamic_mapped


class GrapheneSolidContainer(graphene.Interface):
    solids = non_null_list(GrapheneSolid)

    class Meta:
        name = "SolidContainer"


class GrapheneCompositeSolidDefinition(graphene.ObjectType, ISolidDefinitionMixin):
    solids = non_null_list(GrapheneSolid)
    input_mappings = non_null_list(GrapheneInputMapping)
    output_mappings = non_null_list(GrapheneOutputMapping)

    class Meta:
        interfaces = (GrapheneISolidDefinition, GrapheneSolidContainer)
        name = "CompositeSolidDefinition"

    def __init__(self, represented_pipeline, solid_def_name):
        self._represented_pipeline = check.inst_param(
            represented_pipeline, "represented_pipeline", RepresentedPipeline
        )
        self._solid_def_snap = represented_pipeline.get_solid_def_snap(solid_def_name)
        self._comp_solid_dep_index = represented_pipeline.get_dep_structure_index(solid_def_name)
        super().__init__(name=solid_def_name, description=self._solid_def_snap.description)
        ISolidDefinitionMixin.__init__(self, represented_pipeline, solid_def_name)

    def resolve_solids(self, _graphene_info):
        return build_solids(self._represented_pipeline, self._comp_solid_dep_index)

    def resolve_output_mappings(self, _graphene_info):
        return [
            GrapheneOutputMapping(
                self._represented_pipeline,
                self._comp_solid_dep_index,
                self._solid_def_snap.name,
                output_def_snap.name,
            )
            for output_def_snap in self._solid_def_snap.output_def_snaps
        ]

    def resolve_input_mappings(self, _graphene_info):
        return [
            GrapheneInputMapping(
                self._represented_pipeline,
                self._comp_solid_dep_index,
                self._solid_def_snap.name,
                input_def_snap.name,
            )
            for input_def_snap in self._solid_def_snap.input_def_snaps
        ]


class GrapheneSolidHandle(graphene.ObjectType):
    handleID = graphene.NonNull(graphene.String)
    solid = graphene.NonNull(GrapheneSolid)
    parent = graphene.Field(lambda: GrapheneSolidHandle)

    class Meta:
        name = "SolidHandle"

    def __init__(self, handle, solid, parent):
        super().__init__()
        # FIXME this really seems wrong
        self.handleID = check.inst_param(handle, "handle", SolidHandle)
        self.solid = check.inst_param(solid, "solid", GrapheneSolid)
        self.parent = check.opt_inst_param(parent, "parent", GrapheneSolidHandle)


def build_solid_definition(represented_pipeline, solid_def_name):
    check.inst_param(represented_pipeline, "represented_pipeline", RepresentedPipeline)
    check.str_param(solid_def_name, "solid_def_name")

    solid_def_snap = represented_pipeline.get_solid_def_snap(solid_def_name)

    if isinstance(solid_def_snap, SolidDefSnap):
        return GrapheneSolidDefinition(represented_pipeline, solid_def_snap.name)

    if isinstance(solid_def_snap, CompositeSolidDefSnap):
        return GrapheneCompositeSolidDefinition(represented_pipeline, solid_def_snap.name)

    check.failed("Unknown solid definition type {type}".format(type=type(solid_def_snap)))


types = [
    GrapheneCompositeSolidDefinition,
    GrapheneInput,
    GrapheneInputDefinition,
    GrapheneInputMapping,
    GrapheneISolidDefinition,
    GrapheneOutput,
    GrapheneOutputDefinition,
    GrapheneOutputMapping,
    GrapheneResourceRequirement,
    GrapheneSolid,
    GrapheneSolidContainer,
    GrapheneSolidDefinition,
    GrapheneSolidHandle,
]



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestReexecution.test_full_pipeline_reexecution_fs_storage[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'launchPipelineExecution': {
        '__typename': 'LaunchPipelineRunSuccess',
        'run': {
            'mode': 'default',
            'pipeline': {
                'name': 'csv_hello_world'
            },
            'runConfigYaml': '<runConfigYaml dummy value>',
            'runId': '<runId dummy value>',
            'status': 'STARTING',
            'tags': [
            ]
        }
    }
}

snapshots['TestReexecution.test_full_pipeline_reexecution_fs_storage[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'launchPipelineExecution': {
        '__typename': 'LaunchPipelineRunSuccess',
        'run': {
            'mode': 'default',
            'pipeline': {
                'name': 'csv_hello_world'
            },
            'runConfigYaml': '<runConfigYaml dummy value>',
            'runId': '<runId dummy value>',
            'status': 'STARTING',
            'tags': [
            ]
        }
    }
}

snapshots['TestReexecution.test_full_pipeline_reexecution_in_memory_storage[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = {
    'launchPipelineExecution': {
        '__typename': 'LaunchPipelineRunSuccess',
        'run': {
            'mode': 'default',
            'pipeline': {
                'name': 'csv_hello_world'
            },
            'runConfigYaml': '<runConfigYaml dummy value>',
            'runId': '<runId dummy value>',
            'status': 'STARTING',
            'tags': [
            ]
        }
    }
}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestComputeLogs.test_get_compute_logs_over_graphql[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = {
    'stdout': {
        'data': '''HELLO WORLD
'''
    }
}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestMaterializations.test_materializations[sqlite_with_default_run_launcher_deployed_grpc_env] 1'] = [
    'PipelineStartingEvent',
    'PipelineStartEvent',
    'LogsCapturedEvent',
    'ExecutionStepStartEvent',
    'StepMaterializationEvent',
    'ExecutionStepOutputEvent',
    'HandledOutputEvent',
    'ExecutionStepSuccessEvent',
    'PipelineSuccessEvent'
]

snapshots['TestMaterializations.test_materializations[sqlite_with_default_run_launcher_managed_grpc_env] 1'] = [
    'PipelineStartingEvent',
    'PipelineStartEvent',
    'LogsCapturedEvent',
    'ExecutionStepStartEvent',
    'StepMaterializationEvent',
    'ExecutionStepOutputEvent',
    'HandledOutputEvent',
    'ExecutionStepSuccessEvent',
    'PipelineSuccessEvent'
]



import graphene
from dagster import check
from dagster.core.host_representation import ExternalSensor, SensorSelector
from dagster.core.scheduler.job import JobState
from dagster_graphql.implementation.utils import capture_error, check_read_only

from ..implementation.fetch_sensors import get_sensor_next_tick, start_sensor, stop_sensor
from .errors import (
    GraphenePythonError,
    GrapheneReadOnlyError,
    GrapheneRepositoryNotFoundError,
    GrapheneSensorNotFoundError,
)
from .inputs import GrapheneSensorSelector
from .jobs import GrapheneFutureJobTick, GrapheneJobState
from .util import non_null_list


class GrapheneSensor(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    jobOriginId = graphene.NonNull(graphene.String)
    name = graphene.NonNull(graphene.String)
    pipelineName = graphene.String()
    solidSelection = graphene.List(graphene.String)
    mode = graphene.String()
    sensorState = graphene.NonNull(GrapheneJobState)
    minIntervalSeconds = graphene.NonNull(graphene.Int)
    description = graphene.String()
    nextTick = graphene.Field(GrapheneFutureJobTick)

    class Meta:
        name = "Sensor"

    def __init__(self, graphene_info, external_sensor):
        self._external_sensor = check.inst_param(external_sensor, "external_sensor", ExternalSensor)
        self._sensor_state = graphene_info.context.instance.get_job_state(
            self._external_sensor.get_external_origin_id()
        )

        if not self._sensor_state:
            # Also include a SensorState for a stopped sensor that may not
            # have a stored database row yet
            self._sensor_state = self._external_sensor.get_default_job_state(
                graphene_info.context.instance
            )

        super().__init__(
            name=external_sensor.name,
            jobOriginId=external_sensor.get_external_origin_id(),
            pipelineName=external_sensor.pipeline_name,
            solidSelection=external_sensor.solid_selection,
            mode=external_sensor.mode,
            minIntervalSeconds=external_sensor.min_interval_seconds,
            description=external_sensor.description,
        )

    def resolve_id(self, _):
        return f"{self.name}:{self.pipelineName}" if self.pipelineName else self.name

    def resolve_sensorState(self, _graphene_info):
        return GrapheneJobState(self._sensor_state)

    def resolve_nextTick(self, graphene_info):
        return get_sensor_next_tick(graphene_info, self._sensor_state)


class GrapheneSensorOrError(graphene.Union):
    class Meta:
        types = (
            GrapheneSensor,
            GrapheneSensorNotFoundError,
            GrapheneReadOnlyError,
            GraphenePythonError,
        )
        name = "SensorOrError"


class GrapheneSensors(graphene.ObjectType):
    results = non_null_list(GrapheneSensor)

    class Meta:
        name = "Sensors"


class GrapheneSensorsOrError(graphene.Union):
    class Meta:
        types = (GrapheneSensors, GrapheneRepositoryNotFoundError, GraphenePythonError)
        name = "SensorsOrError"


class GrapheneStartSensorMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneSensorOrError)

    class Arguments:
        sensor_selector = graphene.NonNull(GrapheneSensorSelector)

    class Meta:
        name = "StartSensorMutation"

    @capture_error
    @check_read_only
    def mutate(self, graphene_info, sensor_selector):
        return start_sensor(graphene_info, SensorSelector.from_graphql_input(sensor_selector))


class GrapheneStopSensorMutationResult(graphene.ObjectType):
    jobState = graphene.Field(GrapheneJobState)

    class Meta:
        name = "StopSensorMutationResult"

    def __init__(self, job_state):
        super().__init__()
        self._job_state = check.inst_param(job_state, "job_state", JobState)

    def resolve_jobState(self, _graphene_info):
        if not self._job_state:
            return None

        return GrapheneJobState(job_state=self._job_state)


class GrapheneStopSensorMutationResultOrError(graphene.Union):
    class Meta:
        types = (GrapheneStopSensorMutationResult, GrapheneReadOnlyError, GraphenePythonError)
        name = "StopSensorMutationResultOrError"


class GrapheneStopSensorMutation(graphene.Mutation):
    Output = graphene.NonNull(GrapheneStopSensorMutationResultOrError)

    class Arguments:
        job_origin_id = graphene.NonNull(graphene.String)

    class Meta:
        name = "StopSensorMutation"

    @capture_error
    @check_read_only
    def mutate(self, graphene_info, job_origin_id):
        return stop_sensor(graphene_info, job_origin_id)


types = [
    GrapheneSensor,
    GrapheneSensorOrError,
    GrapheneSensors,
    GrapheneSensorsOrError,
    GrapheneStopSensorMutation,
    GrapheneStopSensorMutationResult,
    GrapheneStopSensorMutationResultOrError,
    GrapheneStopSensorMutation,
]



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['TestSensors.test_get_sensor[non_launchable_sqlite_instance_lazy_repository] 1'] = {
    '__typename': 'Sensor',
    'id': 'always_no_config_sensor:no_config_pipeline',
    'minIntervalSeconds': 30,
    'mode': 'default',
    'name': 'always_no_config_sensor',
    'nextTick': None,
    'pipelineName': 'no_config_pipeline',
    'sensorState': {
        'runs': [
        ],
        'runsCount': 0,
        'status': 'STOPPED',
        'ticks': [
        ]
    },
    'solidSelection': None
}



import tempfile

import pytest
from dagster.core.test_utils import instance_for_test_tempdir

from .setup import define_test_out_of_process_context


@pytest.yield_fixture(scope="function")
def graphql_context():
    with tempfile.TemporaryDirectory() as temp_dir:
        with instance_for_test_tempdir(
            temp_dir,
            overrides={
                "scheduler": {
                    "module": "dagster.utils.test",
                    "class": "FilesystemTestScheduler",
                    "config": {"base_dir": temp_dir},
                }
            },
        ) as instance:
            with define_test_out_of_process_context(instance) as context:
                yield context



from dagster_graphql.test.utils import execute_dagster_graphql, infer_pipeline_selector

from .graphql_context_test_suite import NonLaunchableGraphQLContextTestMatrix
from .setup import csv_hello_world_solids_config

RUN_CONFIG_SCHEMA_QUERY = """
query($selector: PipelineSelector! $mode: String!)
{
  runConfigSchemaOrError(selector: $selector, mode: $mode){
    __typename
    ... on RunConfigSchema {
      rootConfigType {
        key
      }
      allConfigTypes {
        key
      }
    }
  }
}
"""

RUN_CONFIG_SCHEMA_CONFIG_TYPE_QUERY = """
query($selector: PipelineSelector! $mode: String! $configTypeName: String!)
{
  runConfigSchemaOrError(selector: $selector, mode: $mode){
    __typename
    ... on RunConfigSchema {
      configTypeOrError(configTypeName: $configTypeName) {
        __typename
        ... on EnumConfigType {
          name
        }
        ... on RegularConfigType {
          name
        }
        ... on CompositeConfigType {
          name
        }
      }
    }
  }
}
"""


RUN_CONFIG_SCHEMA_CONFIG_VALIDATION_QUERY = """
query PipelineQuery(
    $runConfigData: RunConfigData,
    $selector: PipelineSelector!,
    $mode: String!
) {
  runConfigSchemaOrError(selector: $selector mode: $mode) {
    ... on RunConfigSchema {
      isRunConfigValid(runConfigData: $runConfigData) {
        __typename
        ... on PipelineConfigValidationValid {
            pipelineName
        }
        ... on PipelineConfigValidationInvalid {
            pipelineName
            errors {
                __typename
                ... on RuntimeMismatchConfigError {
                    valueRep
                }
                ... on MissingFieldConfigError {
                    field { name }
                }
                ... on MissingFieldsConfigError {
                    fields { name }
                }
                ... on FieldNotDefinedConfigError {
                    fieldName
                }
                ... on FieldsNotDefinedConfigError {
                    fieldNames
                }
                ... on SelectorTypeConfigError {
                    incomingFields
                }
                message
                reason
                stack {
                    entries {
                        __typename
                        ... on EvaluationStackPathEntry {
                            fieldName
                        }
                        ... on EvaluationStackListItemEntry {
                            listIndex
                        }
                    }
                }
            }
        }
        ... on PipelineNotFoundError {
            pipelineName
        }
      }
    }
  }
}
"""


class TestEnvironmentSchema(NonLaunchableGraphQLContextTestMatrix):
    def test_successful_run_config_schema(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "multi_mode_with_resources")
        result = execute_dagster_graphql(
            graphql_context,
            RUN_CONFIG_SCHEMA_QUERY,
            variables={
                "selector": selector,
                "mode": "add_mode",
            },
        )
        assert result.data["runConfigSchemaOrError"]["__typename"] == "RunConfigSchema"

    def test_run_config_schema_pipeline_not_found(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "jkdjfkdjfd")
        result = execute_dagster_graphql(
            graphql_context,
            RUN_CONFIG_SCHEMA_QUERY,
            variables={"selector": selector, "mode": "add_mode"},
        )
        assert result.data["runConfigSchemaOrError"]["__typename"] == "PipelineNotFoundError"

    def test_run_config_schema_solid_not_found(self, graphql_context):
        selector = infer_pipeline_selector(
            graphql_context, "multi_mode_with_resources", ["kdjfkdj"]
        )
        result = execute_dagster_graphql(
            graphql_context,
            RUN_CONFIG_SCHEMA_QUERY,
            variables={
                "selector": selector,
                "mode": "add_mode",
            },
        )
        assert result.data["runConfigSchemaOrError"]["__typename"] == "InvalidSubsetError"

    def test_run_config_schema_mode_not_found(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "multi_mode_with_resources")
        result = execute_dagster_graphql(
            graphql_context,
            RUN_CONFIG_SCHEMA_QUERY,
            variables={"selector": selector, "mode": "kdjfdk"},
        )
        assert result.data["runConfigSchemaOrError"]["__typename"] == "ModeNotFoundError"

    def test_basic_valid_config_on_run_config_schema(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
        result = execute_dagster_graphql(
            graphql_context,
            RUN_CONFIG_SCHEMA_CONFIG_VALIDATION_QUERY,
            variables={
                "selector": selector,
                "mode": "default",
                "runConfigData": csv_hello_world_solids_config(),
            },
        )

        assert not result.errors
        assert result.data
        assert (
            result.data["runConfigSchemaOrError"]["isRunConfigValid"]["__typename"]
            == "PipelineConfigValidationValid"
        )
        snapshot.assert_match(result.data)

    def test_basic_invalid_config_on_run_config_schema(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, "csv_hello_world")
        result = execute_dagster_graphql(
            graphql_context,
            RUN_CONFIG_SCHEMA_CONFIG_VALIDATION_QUERY,
            variables={
                "selector": selector,
                "mode": "default",
                "runConfigData": {"nope": "kdjfd"},
            },
        )

        assert not result.errors
        assert result.data
        assert (
            result.data["runConfigSchemaOrError"]["isRunConfigValid"]["__typename"]
            == "PipelineConfigValidationInvalid"
        )
        snapshot.assert_match(result.data)
